@@ -1,4 +1,3 @@
-import asyncio
 import base64
 import binascii
 import json
@@ -13,9 +12,9 @@ from multidict import CIMultiDict
 
 from .hdrs import (CONTENT_DISPOSITION, CONTENT_ENCODING, CONTENT_LENGTH,
                    CONTENT_TRANSFER_ENCODING, CONTENT_TYPE)
-from .helpers import CHAR, PY_35, PY_352, TOKEN, parse_mimetype, reify
+from .helpers import CHAR, TOKEN, parse_mimetype, reify
 from .http import HttpParser
-from .payload import (BytesPayload, LookupError, Payload, StringPayload,
+from .payload import (JsonPayload, LookupError, Payload, StringPayload,
                       get_payload, payload_type)
 
 
@@ -161,51 +160,44 @@ def content_disposition_filename(params, name='filename'):
         return value
 
 
-class MultipartResponseWrapper(object):
-    """Wrapper around the :class:`MultipartBodyReader` to take care about
-    underlying connection and close it when it needs in."""
+class MultipartResponseWrapper:
+    """Wrapper around the MultipartBodyReader.
+
+    It takes care about
+    underlying connection and close it when it needs in.
+    """
 
     def __init__(self, resp, stream):
         self.resp = resp
         self.stream = stream
 
-    if PY_35:
-        def __aiter__(self):
-            return self
+    def __aiter__(self):
+        return self
 
-        if not PY_352:  # pragma: no cover
-            __aiter__ = asyncio.coroutine(__aiter__)
-
-        @asyncio.coroutine
-        def __anext__(self):
-            part = yield from self.next()
-            if part is None:
-                raise StopAsyncIteration  # NOQA
-            return part
+    async def __anext__(self):
+        part = await self.next()
+        if part is None:
+            raise StopAsyncIteration  # NOQA
+        return part
 
     def at_eof(self):
-        """Returns ``True`` when all response data had been read.
-
-        :rtype: bool
-        """
+        """Returns True when all response data had been read."""
         return self.resp.content.at_eof()
 
-    @asyncio.coroutine
-    def next(self):
+    async def next(self):
         """Emits next multipart reader object."""
-        item = yield from self.stream.next()
+        item = await self.stream.next()
         if self.stream.at_eof():
-            yield from self.release()
+            await self.release()
         return item
 
-    @asyncio.coroutine
-    def release(self):
+    async def release(self):
         """Releases the connection gracefully, reading all the content
         to the void."""
-        yield from self.resp.release()
+        await self.resp.release()
 
 
-class BodyPartReader(object):
+class BodyPartReader:
     """Multipart reader for single body part."""
 
     chunk_size = 8192
@@ -223,100 +215,77 @@ class BodyPartReader(object):
         self._content_eof = 0
         self._cache = {}
 
-    if PY_35:
-        def __aiter__(self):
-            return self
+    def __aiter__(self):
+        return self
 
-        if not PY_352:  # pragma: no cover
-            __aiter__ = asyncio.coroutine(__aiter__)
+    async def __anext__(self):
+        part = await self.next()
+        if part is None:
+            raise StopAsyncIteration  # NOQA
+        return part
 
-        @asyncio.coroutine
-        def __anext__(self):
-            part = yield from self.next()
-            if part is None:
-                raise StopAsyncIteration  # NOQA
-            return part
-
-    @asyncio.coroutine
-    def next(self):
-        item = yield from self.read()
+    async def next(self):
+        item = await self.read()
         if not item:
             return None
         return item
 
-    @asyncio.coroutine
-    def read(self, *, decode=False):
+    async def read(self, *, decode=False):
         """Reads body part data.
 
-        :param bool decode: Decodes data following by encoding
-                            method from `Content-Encoding` header. If it missed
-                            data remains untouched
-
-        :rtype: bytearray
+        decode: Decodes data following by encoding
+                method from Content-Encoding header. If it missed
+                data remains untouched
         """
         if self._at_eof:
             return b''
         data = bytearray()
         while not self._at_eof:
-            data.extend((yield from self.read_chunk(self.chunk_size)))
+            data.extend((await self.read_chunk(self.chunk_size)))
         if decode:
             return self.decode(data)
         return data
 
-    @asyncio.coroutine
-    def read_chunk(self, size=chunk_size):
+    async def read_chunk(self, size=chunk_size):
         """Reads body part content chunk of the specified size.
 
-        :param int size: chunk size
-
-        :rtype: bytearray
+        size: chunk size
         """
         if self._at_eof:
             return b''
         if self._length:
-            chunk = yield from self._read_chunk_from_length(size)
+            chunk = await self._read_chunk_from_length(size)
         else:
-            chunk = yield from self._read_chunk_from_stream(size)
+            chunk = await self._read_chunk_from_stream(size)
 
         self._read_bytes += len(chunk)
         if self._read_bytes == self._length:
             self._at_eof = True
         if self._at_eof:
-            assert b'\r\n' == (yield from self._content.readline()), \
+            clrf = await self._content.readline()
+            assert b'\r\n' == clrf, \
                 'reader did not read all the data or it is malformed'
         return chunk
 
-    @asyncio.coroutine
-    def _read_chunk_from_length(self, size):
-        """Reads body part content chunk of the specified size.
-        The body part must has `Content-Length` header with proper value.
-
-        :param int size: chunk size
-
-        :rtype: bytearray
-        """
+    async def _read_chunk_from_length(self, size):
+        # Reads body part content chunk of the specified size.
+        # The body part must has Content-Length header with proper value.
         assert self._length is not None, \
             'Content-Length required for chunked read'
         chunk_size = min(size, self._length - self._read_bytes)
-        chunk = yield from self._content.read(chunk_size)
+        chunk = await self._content.read(chunk_size)
         return chunk
 
-    @asyncio.coroutine
-    def _read_chunk_from_stream(self, size):
-        """Reads content chunk of body part with unknown length.
-        The `Content-Length` header for body part is not necessary.
-
-        :param int size: chunk size
-
-        :rtype: bytearray
-        """
+    async def _read_chunk_from_stream(self, size):
+        # Reads content chunk of body part with unknown length.
+        # The Content-Length header for body part is not necessary.
         assert size >= len(self._boundary) + 2, \
             'Chunk size must be greater or equal than boundary length + 2'
         first_chunk = self._prev_chunk is None
         if first_chunk:
-            self._prev_chunk = yield from self._content.read(size)
+            self._prev_chunk = await self._content.read(size)
 
-        chunk = yield from self._content.read(size)
+        chunk = await self._content.read(size)
         self._content_eof += int(self._content.at_eof())
         assert self._content_eof < 3, "Reading after EOF"
         window = self._prev_chunk + chunk
@@ -337,19 +306,15 @@ class BodyPartReader(object):
         self._prev_chunk = chunk
         return result
 
-    @asyncio.coroutine
-    def readline(self):
-        """Reads body part by line by line.
-
-        :rtype: bytearray
-        """
+    async def readline(self):
+        """Reads body part by line by line."""
         if self._at_eof:
             return b''
 
         if self._unread:
             line = self._unread.popleft()
         else:
-            line = yield from self._content.readline()
+            line = await self._content.readline()
 
         if line.startswith(self._boundary):
             # the very last boundary may not come with \r\n,
@@ -363,61 +328,41 @@ class BodyPartReader(object):
                 self._unread.append(line)
                 return b''
         else:
-            next_line = yield from self._content.readline()
+            next_line = await self._content.readline()
             if next_line.startswith(self._boundary):
                 line = line[:-2]  # strip CRLF but only once
             self._unread.append(next_line)
 
         return line
 
-    @asyncio.coroutine
-    def release(self):
-        """Like :meth:`read`, but reads all the data to the void.
-
-        :rtype: None
-        """
+    async def release(self):
+        """Like read(), but reads all the data to the void."""
         if self._at_eof:
             return
         while not self._at_eof:
-            yield from self.read_chunk(self.chunk_size)
+            await self.read_chunk(self.chunk_size)
 
-    @asyncio.coroutine
-    def text(self, *, encoding=None):
-        """Like :meth:`read`, but assumes that body part contains text data.
-
-        :param str encoding: Custom text encoding. Overrides specified
-                             in charset param of `Content-Type` header
-
-        :rtype: str
-        """
-        data = yield from self.read(decode=True)
+    async def text(self, *, encoding=None):
+        """Like read(), but assumes that body part contains text data."""
+        data = await self.read(decode=True)
         # see https://www.w3.org/TR/html5/forms.html#multipart/form-data-encoding-algorithm # NOQA
         # and https://dvcs.w3.org/hg/xhr/raw-file/tip/Overview.html#dom-xmlhttprequest-send # NOQA
         encoding = encoding or self.get_charset(default='utf-8')
         return data.decode(encoding)
 
-    @asyncio.coroutine
-    def json(self, *, encoding=None):
-        """Like :meth:`read`, but assumes that body parts contains JSON data.
-
-        :param str encoding: Custom JSON encoding. Overrides specified
-                             in charset param of `Content-Type` header
-        """
-        data = yield from self.read(decode=True)
+    async def json(self, *, encoding=None):
+        """Like read(), but assumes that body parts contains JSON data."""
+        data = await self.read(decode=True)
         if not data:
             return None
         encoding = encoding or self.get_charset(default='utf-8')
         return json.loads(data.decode(encoding))
 
-    @asyncio.coroutine
-    def form(self, *, encoding=None):
-        """Like :meth:`read`, but assumes that body parts contains form
+    async def form(self, *, encoding=None):
+        """Like read(), but assumes that body parts contains form
         urlencoded data.
-
-        :param str encoding: Custom form encoding. Overrides specified
-                             in charset param of `Content-Type` header
         """
-        data = yield from self.read(decode=True)
+        data = await self.read(decode=True)
         if not data:
             return None
         encoding = encoding or self.get_charset(default='utf-8')
@@ -426,28 +371,12 @@ class BodyPartReader(object):
                          encoding=encoding)
 
     def at_eof(self):
-        """Returns ``True`` if the boundary was reached or
-        ``False`` otherwise.
-
-        :rtype: bool
-        """
+        """Returns True if the boundary was reached or False otherwise."""
         return self._at_eof
 
     def decode(self, data):
-        """Decodes data according the specified `Content-Encoding`
-        or `Content-Transfer-Encoding` headers value.
-
-        Supports ``gzip``, ``deflate`` and ``identity`` encodings for
-        `Content-Encoding` header.
-
-        Supports ``base64``, ``quoted-printable``, ``binary`` encodings for
-        `Content-Transfer-Encoding` header.
-
-        :param bytearray data: Data to decode.
-
-        :raises: :exc:`RuntimeError` - if encoding is unknown.
-
-        :rtype: bytes
+        """Decodes data according the specified Content-Encoding
+        or Content-Transfer-Encoding headers value.
         """
         if CONTENT_TRANSFER_ENCODING in self.headers:
             data = self._decode_content_transfer(data)
@@ -481,24 +410,25 @@ class BodyPartReader(object):
                                ''.format(encoding))
 
     def get_charset(self, default=None):
-        """Returns charset parameter from ``Content-Type`` header or default.
-        """
+        """Returns charset parameter from Content-Type header or default."""
         ctype = self.headers.get(CONTENT_TYPE, '')
-        *_, params = parse_mimetype(ctype)
-        return params.get('charset', default)
+        mimetype = parse_mimetype(ctype)
+        return mimetype.parameters.get('charset', default)
 
     @reify
     def name(self):
-        """Returns filename specified in Content-Disposition header or ``None``
-        if missed or header is malformed."""
+        """Returns name specified in Content-Disposition header or None
+        if missed or header is malformed.
+        """
         _, params = parse_content_disposition(
             self.headers.get(CONTENT_DISPOSITION))
         return content_disposition_filename(params, 'name')
 
     @reify
     def filename(self):
-        """Returns filename specified in Content-Disposition header or ``None``
-        if missed or header is malformed."""
+        """Returns filename specified in Content-Disposition header or None
+        if missed or header is malformed.
+        """
         _, params = parse_content_disposition(
             self.headers.get(CONTENT_DISPOSITION))
         return content_disposition_filename(params, 'filename')
@@ -519,16 +449,15 @@ class BodyPartReaderPayload(Payload):
         if params:
             self.set_content_disposition('attachment', **params)
 
-    @asyncio.coroutine
-    def write(self, writer):
+    async def write(self, writer):
         field = self._value
-        chunk = yield from field.read_chunk(size=2**16)
+        chunk = await field.read_chunk(size=2**16)
         while chunk:
             writer.write(field.decode(chunk))
-            chunk = yield from field.read_chunk(size=2**16)
+            chunk = await field.read_chunk(size=2**16)
 
 
-class MultipartReader(object):
+class MultipartReader:
     """Multipart body reader."""
 
     #: Response wrapper, used when multipart readers constructs from response.
@@ -548,19 +477,14 @@ class MultipartReader(object):
         self._at_bof = True
         self._unread = []
 
-    if PY_35:
-        def __aiter__(self):
-            return self
+    def __aiter__(self):
+        return self
 
-        if not PY_352:  # pragma: no cover
-            __aiter__ = asyncio.coroutine(__aiter__)
-
-        @asyncio.coroutine
-        def __anext__(self):
-            part = yield from self.next()
-            if part is None:
-                raise StopAsyncIteration  # NOQA
-            return part
+    async def __anext__(self):
+        part = await self.next()
+        if part is None:
+            raise StopAsyncIteration  # NOQA
+        return part
 
     @classmethod
     def from_response(cls, response):
@@ -573,43 +497,38 @@ class MultipartReader(object):
         return obj
 
     def at_eof(self):
-        """Returns ``True`` if the final boundary was reached or
-        ``False`` otherwise.
-
-        :rtype: bool
+        """Returns True if the final boundary was reached or
+        False otherwise.
         """
         return self._at_eof
 
-    @asyncio.coroutine
-    def next(self):
+    async def next(self):
         """Emits the next multipart body part."""
         # So, if we're at BOF, we need to skip till the boundary.
         if self._at_eof:
             return
-        yield from self._maybe_release_last_part()
+        await self._maybe_release_last_part()
         if self._at_bof:
-            yield from self._read_until_first_boundary()
+            await self._read_until_first_boundary()
             self._at_bof = False
         else:
-            yield from self._read_boundary()
+            await self._read_boundary()
         if self._at_eof:  # we just read the last boundary, nothing to do there
             return
-        self._last_part = yield from self.fetch_next_part()
+        self._last_part = await self.fetch_next_part()
         return self._last_part
 
-    @asyncio.coroutine
-    def release(self):
+    async def release(self):
         """Reads all the body parts to the void till the final boundary."""
         while not self._at_eof:
-            item = yield from self.next()
+            item = await self.next()
             if item is None:
                 break
-            yield from item.release()
+            await item.release()
 
-    @asyncio.coroutine
-    def fetch_next_part(self):
+    async def fetch_next_part(self):
         """Returns the next body part reader."""
-        headers = yield from self._read_headers()
+        headers = await self._read_headers()
         return self._get_part_reader(headers)
 
     def _get_part_reader(self, headers):
@@ -619,8 +538,9 @@ class MultipartReader(object):
         :param dict headers: Response headers
         """
         ctype = headers.get(CONTENT_TYPE, '')
-        mtype, *_ = parse_mimetype(ctype)
-        if mtype == 'multipart':
+        mimetype = parse_mimetype(ctype)
+
+        if mimetype.type == 'multipart':
             if self.multipart_reader_cls is None:
                 return type(self)(headers, self._content)
             return self.multipart_reader_cls(headers, self._content)
@@ -628,31 +548,31 @@ class MultipartReader(object):
             return self.part_reader_cls(self._boundary, headers, self._content)
 
     def _get_boundary(self):
-        mtype, *_, params = parse_mimetype(self.headers[CONTENT_TYPE])
+        mimetype = parse_mimetype(self.headers[CONTENT_TYPE])
 
-        assert mtype == 'multipart', 'multipart/* content type expected'
+        assert mimetype.type == 'multipart', (
+            'multipart/* content type expected'
+        )
 
-        if 'boundary' not in params:
+        if 'boundary' not in mimetype.parameters:
             raise ValueError('boundary missed for Content-Type: %s'
                              % self.headers[CONTENT_TYPE])
 
-        boundary = params['boundary']
+        boundary = mimetype.parameters['boundary']
         if len(boundary) > 70:
             raise ValueError('boundary %r is too long (70 chars max)'
                              % boundary)
 
         return boundary
 
-    @asyncio.coroutine
-    def _readline(self):
+    async def _readline(self):
         if self._unread:
             return self._unread.pop()
-        return (yield from self._content.readline())
+        return await self._content.readline()
 
-    @asyncio.coroutine
-    def _read_until_first_boundary(self):
+    async def _read_until_first_boundary(self):
         while True:
-            chunk = yield from self._readline()
+            chunk = await self._readline()
             if chunk == b'':
                 raise ValueError("Could not find starting boundary %r"
                                  % (self._boundary))
@@ -663,15 +583,14 @@ class MultipartReader(object):
                 self._at_eof = True
                 return
 
-    @asyncio.coroutine
-    def _read_boundary(self):
-        chunk = (yield from self._readline()).rstrip()
+    async def _read_boundary(self):
+        chunk = (await self._readline()).rstrip()
         if chunk == self._boundary:
             pass
         elif chunk == self._boundary + b'--':
             self._at_eof = True
-            epilogue = yield from self._readline()
-            next_line = yield from self._readline()
+            epilogue = await self._readline()
+            next_line = await self._readline()
 
             # the epilogue is expected and then either the end of input or the
             # parent multipart boundary, if the parent boundary is found then
@@ -688,11 +607,10 @@ class MultipartReader(object):
             raise ValueError('Invalid boundary %r, expected %r'
                              % (chunk, self._boundary))
 
-    @asyncio.coroutine
-    def _read_headers(self):
+    async def _read_headers(self):
         lines = [b'']
         while True:
-            chunk = yield from self._content.readline()
+            chunk = await self._content.readline()
             chunk = chunk.strip()
             lines.append(chunk)
             if not chunk:
@@ -701,12 +619,11 @@ class MultipartReader(object):
         headers, *_ = parser.parse_headers(lines)
         return headers
 
-    @asyncio.coroutine
-    def _maybe_release_last_part(self):
+    async def _maybe_release_last_part(self):
         """Ensures that the last read body part is read completely."""
         if self._last_part is not None:
             if not self._last_part.at_eof():
-                yield from self._last_part.release()
+                await self._last_part.release()
             self._unread.extend(self._last_part._unread)
             self._last_part = None
 
@@ -716,11 +633,18 @@ class MultipartWriter(Payload):
 
     def __init__(self, subtype='mixed', boundary=None):
         boundary = boundary if boundary is not None else uuid.uuid4().hex
+        # The underlying Payload API demands a str (utf-8), not bytes,
+        # so we need to ensure we don't lose anything during conversion.
+        # As a result, require the boundary to be ASCII only.
+        # In both situations.
+
         try:
-            self._boundary = boundary.encode('us-ascii')
+            self._boundary = boundary.encode('ascii')
         except UnicodeEncodeError:
-            raise ValueError('boundary should contains ASCII only chars')
-        ctype = 'multipart/{}; boundary="{}"'.format(subtype, boundary)
+            raise ValueError('boundary should contain ASCII only chars') \
+                from None
+        ctype = ('multipart/{}; boundary={}'
+                 .format(subtype, self._boundary_value))
 
         super().__init__(None, content_type=ctype)
 
@@ -740,9 +664,44 @@ class MultipartWriter(Payload):
     def __len__(self):
         return len(self._parts)
 
+    _valid_tchar_regex = re.compile(br"\A[!#$%&'*+\-.^_`|~\w]+\Z")
+    _invalid_qdtext_char_regex = re.compile(br"[\x00-\x08\x0A-\x1F\x7F]")
+
+    @property
+    def _boundary_value(self):
+        """Wrap boundary parameter value in quotes, if necessary.
+
+        Reads self.boundary and returns a unicode sting.
+        """
+        # Refer to RFCs 7231, 7230, 5234.
+        #
+        # parameter      = token "=" ( token / quoted-string )
+        # token          = 1*tchar
+        # quoted-string  = DQUOTE *( qdtext / quoted-pair ) DQUOTE
+        # qdtext         = HTAB / SP / %x21 / %x23-5B / %x5D-7E / obs-text
+        # obs-text       = %x80-FF
+        # quoted-pair    = "\" ( HTAB / SP / VCHAR / obs-text )
+        # tchar          = "!" / "#" / "$" / "%" / "&" / "'" / "*"
+        #                  / "+" / "-" / "." / "^" / "_" / "`" / "|" / "~"
+        #                  / DIGIT / ALPHA
+        #                  ; any VCHAR, except delimiters
+        # VCHAR           = %x21-7E
+        value = self._boundary
+        if re.match(self._valid_tchar_regex, value):
+            return value.decode('ascii')  # cannot fail
+
+        if re.search(self._invalid_qdtext_char_regex, value):
+            raise ValueError("boundary value contains invalid characters")
+
+        # escape %x5C and %x22
+        quoted_value_content = value.replace(b'\\', b'\\\\')
+        quoted_value_content = quoted_value_content.replace(b'"', b'\\"')
+
+        return '"' + quoted_value_content.decode('ascii') + '"'
+
     @property
     def boundary(self):
-        return self._boundary
+        return self._boundary.decode('ascii')
 
     def append(self, obj, headers=None):
         if headers is None:
@@ -753,10 +712,10 @@ class MultipartWriter(Payload):
                 obj.headers.update(headers)
             else:
                 obj._headers = headers
-            self.append_payload(obj)
+            return self.append_payload(obj)
         else:
             try:
-                self.append_payload(get_payload(obj, headers=headers))
+                return self.append_payload(get_payload(obj, headers=headers))
             except LookupError:
                 raise TypeError
 
@@ -793,16 +752,14 @@ class MultipartWriter(Payload):
         ).encode('utf-8') + b'\r\n'
 
         self._parts.append((payload, headers, encoding, te_encoding))
+        return payload
 
     def append_json(self, obj, headers=None):
         """Helper to append JSON part."""
         if headers is None:
             headers = CIMultiDict()
 
-        data = json.dumps(obj).encode('utf-8')
-        self.append_payload(
-            BytesPayload(
-                data, headers=headers, content_type='application/json'))
+        return self.append_payload(JsonPayload(obj, headers=headers))
 
     def append_form(self, obj, headers=None):
         """Helper to append form urlencoded part."""
@@ -839,15 +796,14 @@ class MultipartWriter(Payload):
         total += 2 + len(self._boundary) + 4  # b'--'+self._boundary+b'--\r\n'
         return total
 
-    @asyncio.coroutine
-    def write(self, writer):
+    async def write(self, writer):
         """Write body."""
         if not self._parts:
             return
 
         for part, headers, encoding, te_encoding in self._parts:
-            yield from writer.write(b'--' + self._boundary + b'\r\n')
-            yield from writer.write(headers)
+            await writer.write(b'--' + self._boundary + b'\r\n')
+            await writer.write(headers)
 
             if encoding or te_encoding:
                 w = MultipartPayloadWriter(writer)
@@ -855,14 +811,14 @@ class MultipartWriter(Payload):
                     w.enable_compression(encoding)
                 if te_encoding:
                     w.enable_encoding(te_encoding)
-                yield from part.write(w)
-                yield from w.write_eof()
+                await part.write(w)
+                await w.write_eof()
             else:
-                yield from part.write(writer)
+                await part.write(writer)
 
-            yield from writer.write(b'\r\n')
+            await writer.write(b'\r\n')
 
-        yield from writer.write(b'--' + self._boundary + b'--\r\n')
+        await writer.write(b'--' + self._boundary + b'--\r\n')
 
 
 class MultipartPayloadWriter:
@@ -884,21 +840,19 @@ class MultipartPayloadWriter:
                      if encoding == 'gzip' else -zlib.MAX_WBITS)
         self._compress = zlib.compressobj(wbits=zlib_mode)
 
-    @asyncio.coroutine
-    def write_eof(self):
+    async def write_eof(self):
         if self._compress is not None:
             chunk = self._compress.flush()
             if chunk:
                 self._compress = None
-                yield from self.write(chunk)
+                await self.write(chunk)
 
         if self._encoding == 'base64':
             if self._encoding_buffer:
-                yield from self._writer.write(base64.b64encode(
+                await self._writer.write(base64.b64encode(
                     self._encoding_buffer))
 
-    @asyncio.coroutine
-    def write(self, chunk):
+    async def write(self, chunk):
         if self._compress is not None:
             if chunk:
                 chunk = self._compress.compress(chunk)
@@ -915,8 +869,8 @@ class MultipartPayloadWriter:
                     buffer[:div * 3], buffer[div * 3:])
                 if enc_chunk:
                     enc_chunk = base64.b64encode(enc_chunk)
-                    yield from self._writer.write(enc_chunk)
+                    await self._writer.write(enc_chunk)
         elif self._encoding == 'quoted-printable':
-            yield from self._writer.write(binascii.b2a_qp(chunk))
+            await self._writer.write(binascii.b2a_qp(chunk))
         else:
-            yield from self._writer.write(chunk)
+            await self._writer.write(chunk)
